package be.iminds.ilabt.jfed.experimenter_gui.bugreporting;

import be.iminds.ilabt.jfed.call_log_output.LogOutput;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetails;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetailsWriter;
import be.iminds.ilabt.jfed.connectivity_tester.ConnectivityTest;
import be.iminds.ilabt.jfed.experimenter_gui.JFedVersionInfo;
import be.iminds.ilabt.jfed.experimenter_gui.connectivity_tester.ConnectivityTester;
import be.iminds.ilabt.jfed.experimenter_gui.util.BugReportExperimentControllerListener;
import be.iminds.ilabt.jfed.highlevel.controller.JavaFXTaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.report.StitchingJobReport;
import be.iminds.ilabt.jfed.highlevel.model.ComponentInfo;
import be.iminds.ilabt.jfed.highlevel.model.SfaModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.log.cache.AllApiCallsList;
import be.iminds.ilabt.jfed.log.cache.ApiCallDetailsRef;
import be.iminds.ilabt.jfed.logging.BugReportAppender;
import be.iminds.ilabt.jfed.logging.JFedLoggerManager;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.ErrorDetails;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.util.KeyUtil;
import ch.qos.logback.classic.spi.ILoggingEvent;
import javafx.beans.property.*;
import javafx.scene.image.Image;
import javanet.staxutils.IndentingXMLEventWriter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.security.cert.CertificateParsingException;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPublicKey;
import java.text.SimpleDateFormat;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.Date;
import java.util.List;
import java.util.Map;

import static be.iminds.ilabt.jfed.experimenter_gui.bugreporting.BugReportConstants.*;
import static ch.qos.logback.classic.spi.ThrowableProxyUtil.asString;
import static java.nio.charset.Charset.defaultCharset;
import static java.nio.file.Files.readAllLines;
import static java.time.Instant.ofEpochMilli;
import static java.time.ZoneId.systemDefault;
import static java.time.ZonedDateTime.ofInstant;
import static java.time.format.DateTimeFormatter.ISO_OFFSET_DATE_TIME;
import static javafx.embed.swing.SwingFXUtils.fromFXImage;
import static javax.imageio.ImageIO.write;
import static javax.xml.stream.XMLEventFactory.newInstance;
import static org.apache.commons.codec.binary.Base64.encodeBase64;

/**
 * This class contains references to all the information that must be included in a bugreport.
 * <p/>
 * It also extends the Task-class, which must be used to write the info to an outputstream. By extending from Task,
 * we are able to display the overall progress to the user.
 * <p/>
 * User: twalcari
 * Date: 6/13/2014
 * Time: 10:35
 */
public class BugReportComposer {
    private final static Logger LOG = LoggerFactory.getLogger(BugReportComposer.class);

    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd");
    private static final int SCREENSHOT_PROGRESS = 100;

    private final JavaFXTaskThread javaFXTaskThread;
    private final GeniUserProvider geniUserProvider;
    private final SfaModel sfaModel;
    private final JFedGuiPreferences jFedPreferences;
    private final JFedLoggerManager jFedLoggerManager;
    private final ConnectivityTester connectivityTester;
    private final BugReportExperimentControllerListener bugReportExperimentControllerListener;
    private final AllApiCallsList allApiCallsList;


    private final StringProperty description = new SimpleStringProperty();
    private final StringProperty email = new SimpleStringProperty();
    private final ReadOnlyStringWrapper version = new ReadOnlyStringWrapper();
    private final ReadOnlyStringWrapper environment = new ReadOnlyStringWrapper();
    private final ReadOnlyStringWrapper credentials = new ReadOnlyStringWrapper();
    private final ReadOnlyStringWrapper certificateExpiration = new ReadOnlyStringWrapper();

    private final ObjectProperty<Image> screenshot = new SimpleObjectProperty<>();


    @Inject
    public BugReportComposer(JavaFXTaskThread javaFXTaskThread, GeniUserProvider geniUserProvider,
                             SfaModel sfaModel, ConnectivityTester connectivityTester,
                             JFedGuiPreferences jFedPreferences, JFedLoggerManager jFedLoggerManager,
                             BugReportExperimentControllerListener bugReportExperimentControllerListener,
                             JFedVersionInfo jFedVersionInfo,
                             AllApiCallsList allApiCallsList) {
        this.javaFXTaskThread = javaFXTaskThread;
        this.geniUserProvider = geniUserProvider;
        this.sfaModel = sfaModel;
        this.connectivityTester = connectivityTester;
        this.jFedPreferences = jFedPreferences;
        this.jFedLoggerManager = jFedLoggerManager;
        this.bugReportExperimentControllerListener = bugReportExperimentControllerListener;
        this.allApiCallsList = allApiCallsList;

        assert javaFXTaskThread != null;
        assert connectivityTester != null;

        //start connectivity tests if this hasn't happened yet
        this.connectivityTester.startTests();


        //now load the info into the disabled textfield
        this.version.set(jFedVersionInfo.getFullVersionString());

        this.environment.set(
                System.getProperty("os.name") + " " + System.getProperty("os.version") + " " + System.getProperty("os.arch")
                        + " - Java " + System.getProperty("java.version")
                        + " (" + System.getProperty("java.vendor") + ")");

        final GeniUser loggedInGeniUser = this.geniUserProvider.getLoggedInGeniUser();

        this.credentials.set(loggedInGeniUser.getUserUrnString());

        //try to get user e-mailaddress
        try {
            for (List<?> list : loggedInGeniUser.getClientCertificateChain().get(0)
                    .getSubjectAlternativeNames()) {
                if (list.get(0).equals(1)) {  //if 1, then this is a rfc822Name-field
                    this.email.set(list.get(1).toString());
                    break;
                }
            }

            Date expirationDate =
                    getEarliestExpire(loggedInGeniUser.getClientCertificateChain());
            certificateExpiration.set(DATE_FORMAT.format(expirationDate));

        } catch (CertificateParsingException e) {
            e.printStackTrace();
        }


    }

    private Date getEarliestExpire(List<X509Certificate> certs) {
        Date res = null;
        for (X509Certificate cert : certs) {
            Date expire = cert.getNotAfter();
            if (res == null || res.after(expire))
                res = expire;
        }
        return res;
    }


    private static final int PROGRESS_OVERALL_INFO = 7;

    private int totalProgress;
    private int progress;
    private boolean finished = false;

    /**
     * Writes a bug-report to the output stream set in the field outputStream. It uses StaX instead of JAXB to limit memory usage.
     * <p/>
     * This function expects the connectivityTests to have completed.
     *
     * @throws java.io.IOException
     */
    public void writeToOutputStream(OutputStream outputStream) throws IOException {
        if (finished)
            throw new IllegalStateException("Composer has already finished");


        if (!isPreparationFinished())
            throw new IllegalStateException("Preparation has not finished!");

        if (outputStream == null)
            throw new IllegalStateException("OutputStream must be set before starting the task!");


        //initialize logreader
        BugReportAppender bugReportAppender = jFedLoggerManager.getBugReportMemoryAppender();

        if (bugReportAppender == null)
            LOG.error("bugReportAppender \"BUGREPORT\" not found! Will not be able to add log entries to bug report.");

        progress = 0;
        //determine the total progress to be made
        totalProgress = PROGRESS_OVERALL_INFO
                + javaFXTaskThread.getCallsCount() * 10
                + (bugReportAppender == null ? 0 : bugReportAppender.getLength())
                + connectivityTester.getTestResults().size()
                + stitchingReportCount() * 50
                + (screenshot.get() != null ? SCREENSHOT_PROGRESS : 0);

        XMLOutputFactory xof = XMLOutputFactory.newInstance();

        try {
            XMLEventWriter xtw = new IndentingXMLEventWriter(xof.createXMLEventWriter(outputStream, "UTF-8"));
            XMLEventFactory eventFactory = newInstance();

            xtw.add(eventFactory.createStartDocument());

            xtw.add(eventFactory.createStartElement(Q_BUGREPORT, null, null));

            //write overall bugreport info to the stream
            xtw.add(eventFactory.createStartElement(Q_DESCRIPTION, null, null));
            xtw.add(eventFactory.createCharacters(description.get()));
            xtw.add(eventFactory.createEndElement(Q_DESCRIPTION, null)); //description

            xtw.add(eventFactory.createStartElement(Q_EMAIL, null, null));
            xtw.add(eventFactory.createCharacters(email.get()));
            xtw.add(eventFactory.createEndElement(Q_EMAIL, null));

            xtw.add(eventFactory.createStartElement(Q_VERSION, null, null));
            xtw.add(eventFactory.createCharacters(version.get()));
            xtw.add(eventFactory.createEndElement(Q_VERSION, null));

            xtw.add(eventFactory.createStartElement(Q_ENVIRONMENT, null, null));
            xtw.add(eventFactory.createCharacters(environment.get()));
            xtw.add(eventFactory.createEndElement(Q_ENVIRONMENT, null));

            xtw.add(eventFactory.createStartElement(Q_REPORTER_CREDENTIAL, null, null));
            xtw.add(eventFactory.createCharacters(credentials.get()));
            xtw.add(eventFactory.createEndElement(Q_REPORTER_CREDENTIAL, null));

            xtw.add(eventFactory.createStartElement(Q_LOGGEDIN_USER_INFO, null, null));
            xtw.add(eventFactory.createCharacters(getLoggedInUserInfo()));
            xtw.add(eventFactory.createEndElement(Q_LOGGEDIN_USER_INFO, null));


            xtw.add(eventFactory.createStartElement(Q_PREFERENCES, null, null));

            if (jFedPreferences.getPropertiesFile() != null) {
                List<String> lines = readAllLines(jFedPreferences.getPropertiesFile().toPath(),
                        defaultCharset());
                for (String line : lines) {
                    xtw.add(eventFactory.createCharacters(line));
                    xtw.add(eventFactory.createCharacters("\n"));
                }
            } else {
                xtw.add(eventFactory.createCharacters("No properties file. (User without home dir?)\n"));
            }

            xtw.add(eventFactory.createEndElement(Q_PREFERENCES, null)); //preferences

            progress += PROGRESS_OVERALL_INFO;

            xtw.flush();


            //write all the calls to the stream
            xtw.add(eventFactory.createStartElement(Q_CALLS, null, null));

//            synchronized (javaFXTaskThread) {
//                for (JavaFXTaskThread.FXTaskExecution taskExecution : javaFXTaskThread.getAllTasks()) {
//                    for (SerializableApiCallDetails details : taskExecution.getApiCallHistory()) {
//
//                        SerializableApiCallDetailsWriter.writeApiCallDetails(details, xtw, eventFactory);
//                        xtw.flush();
//                    }
//
//                    progress += 10;
//                }
//            }
            synchronized (javaFXTaskThread) {
                List<ApiCallDetailsRef> allRefs = allApiCallsList.getCalls();
                for (ApiCallDetailsRef ref : allRefs) {
                    SerializableApiCallDetails details = allApiCallsList.getCache().getApiCallDetails(ref);
                    SerializableApiCallDetailsWriter.writeApiCallDetails(details, xtw, eventFactory);
                    xtw.flush();

                    progress += 10;
                }
            }

            xtw.add(eventFactory.createEndElement(Q_CALLS, null)); //calls

            xtw.add(eventFactory.createStartElement(Q_STITCHING_REPORTS, null, null));

            synchronized (bugReportExperimentControllerListener.getStitchingJobReports()) {
                for (StitchingJobReport stitchingReport : bugReportExperimentControllerListener.getStitchingJobReports()) {
                    writeJobReport(new SerializableJobReport(stitchingReport), xtw, eventFactory);
                    xtw.flush();

                    progress += 50;
                }
            }
            xtw.add(eventFactory.createEndElement(Q_STITCHING_REPORTS, null)); //stitchingreports


            //write all log-lines to the stream
            xtw.add(eventFactory.createStartElement(Q_LOG, null, null));

            if (bugReportAppender != null) {
                List<ILoggingEvent> loggingEvents = bugReportAppender.getListCopy();
                for (ILoggingEvent loggingEvent : loggingEvents) {

                    writeLogLine(loggingEvent, xtw, eventFactory);
                    progress++;

                }
            }


            xtw.add(eventFactory.createEndElement(Q_LOG, null)); //log
            xtw.flush();

            //write all connectivity tests to the stream
            xtw.add(eventFactory.createStartElement(Q_CONNECTIVITY_TESTS, null, null));
            for (Map.Entry<ConnectivityTest, ConnectivityTest.ConnectivityTestResult> entry : connectivityTester.getTestResults().entrySet()) {
                writeConnectivityTest(entry.getKey(), entry.getValue(), xtw, eventFactory);

                progress++;
            }

            xtw.add(eventFactory.createEndElement(Q_CONNECTIVITY_TESTS, null));  //connectivity-tests
            xtw.flush();

            //write current slice information to the stream

            xtw.add(eventFactory.createStartElement(Q_SLICES, null, null));


            for (Slice slice : sfaModel.getSlices()) {
                writeSlice(slice, xtw, eventFactory);

                progress++;
            }


            xtw.add(eventFactory.createEndElement(Q_SLICES, null)); //slices

            xtw.add(eventFactory.createStartElement(Q_SCREENSHOT, null, null));

            if (screenshot.get() != null) {

                try {
                    ByteArrayOutputStream screenshotBuffer = new ByteArrayOutputStream();
                    write(fromFXImage(screenshot.get(), null), "png", screenshotBuffer);
                    byte[] encoded = encodeBase64(screenshotBuffer.toByteArray());
                    xtw.add(eventFactory.createCharacters(new String(encoded)));
                } catch (IOException ex) {
                    LOG.warn("Error while writing bugreport screenshot: ignoring!", ex);
                }

                progress += SCREENSHOT_PROGRESS;
            }


            xtw.add(eventFactory.createEndElement(Q_SCREENSHOT, null)); //screenshot

            xtw.add(eventFactory.createEndElement(Q_BUGREPORT, null)); //bugreport
            xtw.flush();

        } catch (XMLStreamException ex) {
            throw new IOException("Error while writing bug report to stream", ex);
        }

        finished = true;
    }

    private void writeJobReport(SerializableJobReport jobReport, XMLEventWriter xtw, XMLEventFactory eventFactory) throws XMLStreamException {
        xtw.add(eventFactory.createStartElement(Q_JOB_REPORT, null, null));

        xtw.add(eventFactory.createAttribute(Q_TITLE, jobReport.getName()));

        xtw.add(eventFactory.createStartElement(Q_STATES, null, null));
        for (SerializableJobStateReport jobStateReport : jobReport.getJobStateReports()) {
            writeJobStateReport(jobStateReport, xtw, eventFactory);
        }
        xtw.add(eventFactory.createEndElement(Q_STATES, null));

        xtw.add(eventFactory.createEndElement(Q_JOB_REPORT, null));
    }

    private void writeJobStateReport(SerializableJobStateReport jobStateReport, XMLEventWriter xtw, XMLEventFactory eventFactory)
            throws XMLStreamException {
        xtw.add(eventFactory.createStartElement(Q_STATE, null, null));

        xtw.add(eventFactory.createStartElement(Q_NAME, null, null));
        xtw.add(eventFactory.createCharacters(jobStateReport.getName()));
        xtw.add(eventFactory.createEndElement(Q_NAME, null));

        if (jobStateReport.getStart() != null) {
            xtw.add(eventFactory.createStartElement(Q_STATESTATE, null, null));
            xtw.add(eventFactory.createCharacters(jobStateReport.getStatus().name()));
            xtw.add(eventFactory.createEndElement(Q_STATESTATE, null));
        }

        xtw.add(eventFactory.createStartElement(Q_START, null, null));
        xtw.add(eventFactory.createCharacters(ISO_OFFSET_DATE_TIME.format(
                ofInstant(jobStateReport.getStart(), systemDefault()))));
        xtw.add(eventFactory.createEndElement(Q_START, null));

        xtw.add(eventFactory.createStartElement(Q_END, null, null));
        if (jobStateReport.getEnd() != null) {
            xtw.add(eventFactory.createCharacters(ISO_OFFSET_DATE_TIME.format(
                    ofInstant(jobStateReport.getEnd(), systemDefault()))));
        }
        xtw.add(eventFactory.createEndElement(Q_END, null));

        xtw.add(eventFactory.createStartElement(Q_LOGENTRIES, null, null));
        for (LogOutput.LogEntry logEntry : jobStateReport.getLogEntries()) {
            writeLogEntry(logEntry, xtw, eventFactory);
        }
        xtw.add(eventFactory.createEndElement(Q_LOGENTRIES, null)); //logentries


        //write all the taskExecutions and calls to the stream
        xtw.add(eventFactory.createStartElement(Q_TASK_EXECUTIONS, null, null));
        for (SerializableTaskExecution taskExecution : jobStateReport.getTaskExecutions()) {
            writeTaskExecution(xtw, eventFactory, taskExecution);
        }
        xtw.add(eventFactory.createEndElement(Q_TASK_EXECUTIONS, null)); //calls


        xtw.add(eventFactory.createEndElement(Q_STATE, null));   //state
        xtw.flush();
    }

    private void writeTaskExecution(XMLEventWriter xtw, XMLEventFactory eventFactory, SerializableTaskExecution taskExecution) throws XMLStreamException {
        xtw.add(eventFactory.createStartElement(Q_TASK_EXECUTION, null, null));
        xtw.add(eventFactory.createAttribute(Q_ID, taskExecution.getId()));
        xtw.add(eventFactory.createAttribute(Q_NAME, taskExecution.getName()));
        xtw.add(eventFactory.createAttribute(Q_STATUS, taskExecution.getState().name()));
        xtw.add(eventFactory.createAttribute(Q_TASK_CLASS, taskExecution.getTaskClass()));

        if (taskExecution.getStart() != null)
            xtw.add(eventFactory.createAttribute(Q_START, ISO_OFFSET_DATE_TIME.format(
                    ofInstant(taskExecution.getStart(), systemDefault()))));
        if (taskExecution.getEnd() != null)
            xtw.add(eventFactory.createAttribute(Q_END, ISO_OFFSET_DATE_TIME.format(
                    ofInstant(taskExecution.getEnd(), systemDefault()))));

        for (ApiCallDetailsRef apiCallDetailsRef : taskExecution.getApiCallDetails()) {
            xtw.add(eventFactory.createStartElement(Q_CALL_REF, null, null));
            xtw.add(eventFactory.createAttribute(Q_CALL_REF_ID, Integer.toString(apiCallDetailsRef.getApiCallDetailsId())));
            xtw.add(eventFactory.createEndElement(Q_CALL_REF, null));
        }

        xtw.add(eventFactory.createEndElement(Q_TASK_EXECUTION, null));
    }

    private void writeLogEntry(LogOutput.LogEntry logEntry, XMLEventWriter xtw,
                               XMLEventFactory eventFactory) throws XMLStreamException {
        xtw.add(eventFactory.createStartElement(Q_LOGENTRY, null, null));

        xtw.add(eventFactory.createStartElement(Q_TYPE, null, null));
        xtw.add(eventFactory.createCharacters(logEntry.getType().name()));
        xtw.add(eventFactory.createEndElement(Q_TYPE, null));

        xtw.add(eventFactory.createStartElement(Q_TEXT, null, null));
        xtw.add(eventFactory.createCharacters(logEntry.getText()));
        xtw.add(eventFactory.createEndElement(Q_TEXT, null));

        xtw.add(eventFactory.createStartElement(Q_TIME_IN_MS, null, null));
        xtw.add(eventFactory.createCharacters(Long.toString(logEntry.getTimeMs())));
        xtw.add(eventFactory.createEndElement(Q_TIME_IN_MS, null));

        if (logEntry.getExceptionStackTrace() != null) {
            xtw.add(eventFactory.createStartElement(Q_EXCEPTION, null, null));
            xtw.add(eventFactory.createCharacters(logEntry.getExceptionStackTrace()));
            xtw.add(eventFactory.createEndElement(Q_EXCEPTION, null));
        }

        xtw.add(eventFactory.createEndElement(Q_LOGENTRY, null)); //logentry
    }

    private void writeSlice(Slice slice, XMLEventWriter xtw, XMLEventFactory eventFactory) throws XMLStreamException {
        xtw.add(eventFactory.createStartElement(Q_SLICE, null, null));

        xtw.add(eventFactory.createStartElement(Q_SLICE_URN, null, null));
        xtw.add(eventFactory.createCharacters(slice.getUrnString()));
        xtw.add(eventFactory.createEndElement(Q_SLICE_URN, null));

        xtw.add(eventFactory.createStartElement(Q_EXPIRATION_DATE, null, null));
        if (slice.getExpirationDate() != null) {
            xtw.add(eventFactory.createCharacters(DateTimeFormatter.ISO_OFFSET_DATE_TIME.format(slice.getExpirationDate().atZone(ZoneId.systemDefault()))));
        }
        xtw.add(eventFactory.createEndElement(Q_EXPIRATION_DATE, null));

        xtw.add(eventFactory.createStartElement(Q_CURRENT_STATUS, null, null));
        xtw.add(eventFactory.createCharacters(slice.getStatus().detailedToString()));
        xtw.add(eventFactory.createEndElement(Q_CURRENT_STATUS, null));


        for (Sliver sliver : slice.getSlivers()) {
            xtw.add(eventFactory.createStartElement(Q_SLIVER, null, null));

            xtw.add(eventFactory.createStartElement(Q_SLIVER_URN, null, null));
            String sliverUrn = sliver.getUrn() != null ?
                    sliver.getUrn().toString() :
                    "SLIVER-ON: " + sliver.getServer().getDefaultComponentManagerUrn();

            xtw.add(eventFactory.createCharacters(sliverUrn));
            xtw.add(eventFactory.createEndElement(Q_SLIVER_URN, null)); //sliverUrn

            xtw.add(eventFactory.createStartElement(Q_EXPIRATION_DATE, null, null));
            if (sliver.getExpirationDate() != null)
                xtw.add(eventFactory.createCharacters(DateTimeFormatter.ISO_OFFSET_DATE_TIME.format(sliver.getExpirationDate().atZone(ZoneId.systemDefault()))));
            xtw.add(eventFactory.createEndElement(Q_EXPIRATION_DATE, null));

            xtw.add(eventFactory.createStartElement(Q_CURRENT_STATUS, null, null));
            xtw.add(eventFactory.createCharacters(sliver.getStatusString()));
            xtw.add(eventFactory.createEndElement(Q_CURRENT_STATUS, null));

            xtw.add(eventFactory.createStartElement(Q_HAS_MANIFEST, null, null));
            xtw.add(eventFactory.createCharacters(Boolean.toString(sliver.getManifestRspec() != null)));
            xtw.add(eventFactory.createEndElement(Q_HAS_MANIFEST, null));

            xtw.add(eventFactory.createStartElement(Q_COMPONENTS, null, null));
            for (ComponentInfo ci : sliver.getComponentInfo()) {
                if (ci.getClientId() != null || ci.getComponentId() != null) {
                    xtw.add(eventFactory.createStartElement(Q_COMPONENT, null, null));
                    if (ci.getClientId() == null)
                        xtw.add(eventFactory.createCharacters(ci.getClientId()));
                    else
                        xtw.add(eventFactory.createCharacters(ci.getComponentId()));
                    xtw.add(eventFactory.createEndElement(Q_COMPONENT, null)); //component
                }
            }
            xtw.add(eventFactory.createEndElement(Q_COMPONENTS, null)); //components

            if (sliver.getStatus() != null) {
                xtw.add(eventFactory.createStartElement(Q_SLIVER_STATUS, null, null));
                xtw.add(eventFactory.createCharacters(sliver.getStatus().getGlobalStatus().toString()));
                xtw.add(eventFactory.createEndElement(Q_SLIVER_STATUS, null));

                for (ErrorDetails errorDetails : sliver.getStatus().getAllErrors()) {
                    xtw.add(eventFactory.createStartElement(Q_ERROR, null, null));
                    xtw.add(eventFactory.createAttribute(Q_TIMESTAMP,
                            ISO_OFFSET_DATE_TIME.format(ofInstant(
                                    ofEpochMilli(errorDetails.getTimestamp()), systemDefault()))));
                    if (errorDetails.getSliverUrn() != null)
                        xtw.add(eventFactory.createAttribute(Q_SLIVER_URN, errorDetails.getSliverUrn().toString()));
                    if (errorDetails.getComponentUrn() != null)
                        xtw.add(eventFactory.createAttribute(Q_COMPONENT_URN, errorDetails.getComponentUrn().toString()));
                    xtw.add(eventFactory.createCharacters(errorDetails.getError()));
                    xtw.add(eventFactory.createEndElement(Q_ERROR, null)); //error
                }
            }

            xtw.add(eventFactory.createEndElement(Q_SLIVER, null)); //sliver
        }

        xtw.add(eventFactory.createEndElement(Q_SLICE, null));//slice
    }

    private void writeConnectivityTest(ConnectivityTest test, ConnectivityTest.ConnectivityTestResult result, XMLEventWriter xtw, XMLEventFactory eventFactory) throws XMLStreamException {
        xtw.add(eventFactory.createStartElement(Q_CONNECTIVITY_TEST, null, null));

        xtw.add(eventFactory.createStartElement(Q_NAME, null, null));
        xtw.add(eventFactory.createCharacters(test.getName()));
        xtw.add(eventFactory.createEndElement(Q_NAME, null));

        if (result != null) {
            xtw.add(eventFactory.createStartElement(Q_STATUS, null, null));
            xtw.add(eventFactory.createCharacters(result.getStatus().toString()));
            xtw.add(eventFactory.createEndElement(Q_STATUS, null));

            xtw.add(eventFactory.createStartElement(Q_MESSAGE, null, null));
            xtw.add(eventFactory.createCharacters(result.getMessage()));
            xtw.add(eventFactory.createEndElement(Q_MESSAGE, null));

            if (result.getExceptionString() != null) {
                xtw.add(eventFactory.createStartElement(Q_EXCEPTION, null, null));
                xtw.add(eventFactory.createCharacters(result.getExceptionString()));
                xtw.add(eventFactory.createEndElement(Q_EXCEPTION, null));
            }
        }

        xtw.add(eventFactory.createEndElement(Q_CONNECTIVITY_TEST, null)); //connectivityTests
    }

    private void writeLogLine(ILoggingEvent loggingEvent, XMLEventWriter xtw, XMLEventFactory eventFactory) throws XMLStreamException {
        xtw.add(eventFactory.createStartElement(Q_LOGLINE, null, null));

        xtw.add(eventFactory.createStartElement(Q_LEVEL, null, null));
        xtw.add(eventFactory.createCharacters(loggingEvent.getLevel().toString()));
        xtw.add(eventFactory.createEndElement(Q_LEVEL, null));

        xtw.add(eventFactory.createStartElement(Q_CLASSNAME, null, null));
        xtw.add(eventFactory.createCharacters(loggingEvent.getLoggerName()));
        xtw.add(eventFactory.createEndElement(Q_CLASSNAME, null));

        xtw.add(eventFactory.createStartElement(Q_DATETIME, null, null));
        xtw.add(eventFactory.createCharacters(ISO_OFFSET_DATE_TIME.format(ofInstant(ofEpochMilli(loggingEvent.getTimeStamp()), systemDefault()))));
        xtw.add(eventFactory.createEndElement(Q_DATETIME, null));


        xtw.add(eventFactory.createStartElement(Q_TIME, null, null));
        xtw.add(eventFactory.createCharacters(Long.toString(loggingEvent.getTimeStamp())));
        xtw.add(eventFactory.createEndElement(Q_TIME, null));

        xtw.add(eventFactory.createStartElement(Q_THREADNAME, null, null));
        xtw.add(eventFactory.createCharacters(loggingEvent.getThreadName()));
        xtw.add(eventFactory.createEndElement(Q_THREADNAME, null));

        xtw.add(eventFactory.createStartElement(Q_MESSAGE, null, null));
        xtw.add(eventFactory.createCharacters(loggingEvent.getFormattedMessage()));
        xtw.add(eventFactory.createEndElement(Q_MESSAGE, null));

        if (loggingEvent.getThrowableProxy() != null) {
            xtw.add(eventFactory.createStartElement(Q_EXCEPTION, null, null));
            xtw.add(eventFactory.createCharacters(asString(loggingEvent.getThrowableProxy())));
            xtw.add(eventFactory.createEndElement(Q_EXCEPTION, null));
        }

        xtw.add(eventFactory.createEndElement(Q_LOGLINE, null)); //logline
    }

    public void reset() {
        progress = 0;
        totalProgress = 0;
        finished = false;
    }

    public boolean isFinished() {
        return finished;
    }

    public int getTotalProgress() {
        return totalProgress;
    }

    public int getProgress() {
        return progress;
    }

    public boolean isPreparationFinished() {
        return connectivityTester.getTestsFinished();

    }

    private String getLoggedInUserInfo() {
        String userInfoStr = null;
        try {
            if (geniUserProvider.isUserLoggedIn()) {
                GeniUser u = geniUserProvider.getLoggedInGeniUser();
                //TODO: certInfo would better contain info about the certificate such as all data that can be extracted.
                //      PEM does contain all that info but is not human readable.
                userInfoStr = u.getUserUrnString() + " @ " + u.getUserAuthorityServer().getDefaultComponentManagerUrn();
                if (u.getPrivateKey() == null)
                    userInfoStr += " WARNING, no private key!";
                try {
                    String pubkey =
                            KeyUtil.rsaPublicKeyToOpenSshAuthorizedKeysFormatData((RSAPublicKey) u.getClientCertificateChain().get(0).getPublicKey());
                    userInfoStr += " with pubkey " + pubkey;
                } catch (Exception | AssertionError e) {
                    userInfoStr += " ERROR getting pubkey " + e.getMessage();
                }
                try {
                    String certInfo = KeyUtil.x509certificateChainToPem(u.getClientCertificateChain());
                    userInfoStr += " with cert " + certInfo;
                } catch (Exception | AssertionError e) {
                    userInfoStr += " ERROR getting cert " + e.getMessage();
                }
            } else
                userInfoStr = "No user logged in";
        } catch (AssertionError e) {
            if (userInfoStr == null)
                userInfoStr = "NULL then ";
            userInfoStr += " AssertionError getting user info " + e.getMessage();
        } catch (Exception e) {
            if (userInfoStr == null)
                userInfoStr = "NULL then ";
            userInfoStr += " Exception getting user info " + e.getMessage();
        }
        return userInfoStr;
    }

    public String getDescription() {
        return description.get();
    }

    public StringProperty descriptionProperty() {
        return description;
    }

    public void setDescription(String description) {
        this.description.set(description);
    }

    public String getEmail() {
        return email.get();
    }

    public StringProperty emailProperty() {
        return email;
    }

    public void setEmail(String email) {
        this.email.set(email);
    }

    public String getVersion() {
        return version.get();
    }

    public ReadOnlyStringProperty versionProperty() {
        return version.getReadOnlyProperty();
    }

    public String getEnvironment() {
        return environment.get();
    }

    public ReadOnlyStringProperty environmentProperty() {
        return environment.getReadOnlyProperty();
    }


    public String getCredentials() {
        return credentials.get();
    }

    public ReadOnlyStringProperty credentialsProperty() {
        return credentials.getReadOnlyProperty();
    }

    public String getCertificateExpiration() {
        return certificateExpiration.get();
    }

    public ReadOnlyStringProperty certificateExpirationProperty() {
        return certificateExpiration.getReadOnlyProperty();
    }

    public ReadOnlyIntegerProperty callsCountProperty() {
        return javaFXTaskThread.callsCountProperty();
    }

    public int stitchingReportCount() {
        return bugReportExperimentControllerListener.getStitchingJobReports().size();
    }

    public Image getScreenshot() {
        return screenshot.get();
    }

    public ObjectProperty<Image> screenshotProperty() {
        return screenshot;
    }

    public void setScreenshot(Image screenshot) {
        this.screenshot.set(screenshot);
    }
}
