package be.iminds.ilabt.jfed.highlevel.tasks;

import be.iminds.ilabt.jfed.highlevel.controller.JFedHighLevelException;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.model.SfaModel;
import be.iminds.ilabt.jfed.highlevel.model.SfaModelAbstractListener;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.highlevel.util.SliceRegistryUtil;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractGeniAggregateManager;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.*;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 6/17/2015
 * Time: 11:50
 */
public final class CreateSliverTask extends CreateTask {
    private static final Logger LOG = LoggerFactory.getLogger(CreateSliverTask.class);
    //input parameters
    private final Slice slice;
    private final Instant expirationTime;
    private final List<UserSpec> userSpecs;

    @Nullable
    private final RequestRspecSource overridingRequestRspecSource;

    //helper fields
    private final Server auth;

    //results
    private String manifestRspec;
    private Set<Sliver> createdSlivers = null;

    private final List<Task> extraDependencies = new ArrayList<>();

    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    private final SliceRegistryUtil sliceRegistryUtil;
    private final SfaModel sfaModel;
    private final HighLevelTaskFactory highLevelTaskFactory;
    private AbstractGeniAggregateManager.AggregateManagerReply rawReply;


    /**
     * Use this constructor to create a sliver that will be used immediately!
     */
    public CreateSliverTask(Slice slice, Instant expirationTime, Server auth, List<UserSpec> userSpecs,
                            @Nullable RequestRspecSource overridingRequestRspecSource,
                            AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                            SliceRegistryUtil sliceRegistryUtil, SfaModel sfaModel,
                            HighLevelTaskFactory highLevelTaskFactory) {
        super("Create Sliver @ " + auth.getDefaultComponentManagerUrn());
        this.slice = slice;
        this.expirationTime = expirationTime;
        this.auth = auth;
        this.userSpecs = userSpecs;
        this.overridingRequestRspecSource = overridingRequestRspecSource;

        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.sliceRegistryUtil = sliceRegistryUtil;
        this.sfaModel = sfaModel;
        this.highLevelTaskFactory = highLevelTaskFactory;
    }

    @Override
    public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
        // Preparation: get a list of existing slivers on the given authority for the given slice.
        // This allows us to see which slivers were created during the CreateSliver-call

        List<Sliver> existingSlivers = slice.findSlivers(auth);

        FXModelRspec.RequestRspecSpecialCases rspecSpecialCase = FXModelRspec.RequestRspecSpecialCases.NONE;
        if ((auth.isServerType("planetlab"))
                || auth.getDefaultComponentManagerUrn().startsWith("urn:publicid:IDN+ple:ibbtple+authority+")) {
            rspecSpecialCase = FXModelRspec.RequestRspecSpecialCases.PLE;
        }

        AggregateManagerWrapper amWrapper =
                aggregateManagerWrapperFactory.getAggregateManagerWrapper(taskExecution, auth);

        assert !amWrapper.hasSeperateAllocateAndProvision() :
                "AM Wrapper supports provisioning, so we shouldn't be doing CreateSliver!";

        RequestRspecSource requestRspecSource =
                overridingRequestRspecSource != null ?
                        overridingRequestRspecSource : slice.getRequestRspec();
        LOG.trace("HLC.CreateSliverTask request RSpec before modification: expires=" + requestRspecSource.getModelRspec().getExpiresString());
        requestRspecSource = updateExpires(auth, requestRspecSource, expirationTime);
        LOG.debug("HLC.CreateSliverTask request RSpec after modification: expires=" + requestRspecSource.getModelRspec().getExpiresString());
        String rspec = requestRspecSource.getRspecXmlString(rspecSpecialCase);


        LOG.info("createSlivers @ {}  with user SSH keys: {}", auth.getDefaultComponentManagerUrn(), sfaModel.getUserKeys()); //missing here
        try {
            //don't set date here if CreateSliver doesn't set it itself
            Instant usedExpirationDate =
                    amWrapper.isExirationdateSetByCreateSliver()
                            || auth.hasFlag(Server.Flag.workaroundForceRenewAfterCreateSliver)
                            ? expirationTime : null;

            this.manifestRspec = amWrapper.createSliver(
                    slice.getUrn(),
                    slice.getCredentials(),
                    rspec,
                    userSpecs,
                    usedExpirationDate != null ? Date.from(usedExpirationDate) : null
            );
            rawReply = amWrapper.getLastReply();
        } catch (AggregateManagerWrapper.NotEnoughFreeResourcesException e) {
            failedDueToNotEnoughFreeResources = true;
            this.manifestRspec = null;
            throw new JFedHighLevelException("Creation of the sliver failed: could not map to resources " +
                    "(resource shortage or unsupported topology)", e);
        }

        //Note: listener that handles createSliver should update the list of component managers, as it will have changed
        if (this.manifestRspec == null)
            throw new JFedHighLevelException("Creation of the sliver failed.");

        createdSlivers = new HashSet<>();
        Collection<Sliver> manifestSlivers =
                SfaModelAbstractListener.findSliversInManifest(sfaModel, auth, slice.getUrn(), new ManifestRspecSource(manifestRspec, ModelRspecType.FX));
        if (manifestSlivers != null) createdSlivers.addAll(manifestSlivers);

        //add slivers found by sfaModel due to SliverStatus and Status calls. //AL2S is one case where this is neeed, see FEDIBBTDEV-1392
        createdSlivers.addAll(slice.getSlivers().stream().filter(auth::equals).collect(Collectors.toSet()));

        //we don't want to process slivers that already existed!
        createdSlivers.removeAll(existingSlivers);

        if (!createdSlivers.isEmpty())
            sliceRegistryUtil.registerSliversAtSATaskExecution(createdSlivers, this);

        if (createdSlivers.isEmpty())
            throw new JFedHighLevelException("Could not find any slivers that were newly created! " +
                    "Pre-existing slivers: " + existingSlivers.stream().map(Sliver::getUrnString).collect(Collectors.joining(", ")));

    }

    @Override
    public List<Task> initDependsOn() {
        List<Task> res = new ArrayList<>();
        res.add(highLevelTaskFactory.getSliceCredential(slice));
        res.addAll(extraDependencies);
        return res;
    }

    public void addExtraDependency(Task t) {
        extraDependencies.add(t);
    }

    public Slice getSlice() {
        return slice;
    }

    public Server getServer() {
        return auth;
    }


    @Override
    public Set<Sliver> getCreatedSlivers() {
        return Collections.unmodifiableSet(createdSlivers);
    }

    public AbstractGeniAggregateManager.AggregateManagerReply getRawReply() {
        return rawReply;
    }
}
