package be.iminds.ilabt.jfed.lowlevel.api;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * AbstractGeniClearingHouseApi
 */
public abstract class AbstractFederationApi extends AbstractApi {
    private static org.slf4j.Logger LOG = LoggerFactory.getLogger(AbstractFederationApi.class);
    /*
     * How to handle errors due to malformed replies
     * if true, will try to handle errors as much as possible.
     * if false, will throw errors on malformed replies.
     *
     *  true is useful when creating a client tool that needs to be robust
     *  false is useful when creating a test
     * */
    protected boolean handleMalformedReplies = true;

    protected void handleErrorProcessingArguments(XMLRPCCallDetailsWithCodeValueError res, String methodJavaName, String methodGeniName, SfaConnection con, Throwable t) throws JFedException {
        boolean isSuccess = true;
        Object resultCode = res.getResultCode();
//        if (resultCode != null && resultCode instanceof Hashtable) {
//            Hashtable codeStruct = (Hashtable) resultCode;
//            int code = (Integer) codeStruct.get("geni_code");
//            GeniResponseCode genicode = GeniAMResponseCode.getByCode(code);
//            isSuccess = genicode.isSuccess();
//        }
        if (resultCode != null && resultCode instanceof Integer) {
            int code = (Integer) resultCode;
            try {
                GeniResponseCode genicode = GeniUniformFederationApiResponseCode.getByCode(code);
                isSuccess = genicode.isSuccess();
            } catch (Exception e) { /* ignore */ isSuccess = false; }
//            isSuccess = code == 0;
        }

        if (!handleMalformedReplies && isSuccess) {
            //If call failed, error may be ignored.
            log(res, null, methodJavaName, methodGeniName, con, null);
            throw new JFedException("Error parsing "+methodGeniName+" reply: "+(t == null ? "null" : t.getMessage()), t, res);
        }
        else {
            LOG.error("Error parsing " + methodGeniName + " reply: " + t, t);
        }
    }

    @Override
    protected boolean isPossibleTemporaryErrorReply(be.iminds.ilabt.jfed.lowlevel.XMLRPCCallDetails res) {
        if (res instanceof XMLRPCCallDetailsWithCodeValueError) {
            XMLRPCCallDetailsWithCodeValueError geniDetails = (XMLRPCCallDetailsWithCodeValueError) res;
            if (!(geniDetails.getResultCode() instanceof Hashtable))
                return false;

            Hashtable codeStruct = (Hashtable) geniDetails.getResultCode();

            if (codeStruct == null || !codeStruct.containsKey("geni_code"))
                return false;

            int code = (Integer) codeStruct.get("geni_code");

            if (code == 3 &&
                    geniDetails.getResultOutput().trim().equals("Credential owner does not match speaksfor target")) {
                LOG.warn("Received \"Credential owner does not match speaksfor target\" error. Will handle this as \"possible temporary error\" and try again.");
                return true;
            }
        }
        return false;
    }

    public static class GetVersionResult {
        protected String version;
        public static class CredentialType {
            private final String typeName;
            private final String version;

            public CredentialType(String typeName, String version) {
                this.typeName = typeName;
                this.version = version;
            }

            public String getTypeName() { return typeName; }
            public String getVersion() { return version; }
        }
        protected List<CredentialType> supportedCredentialTypes;
        protected Map<String, FieldInfo> fieldInfos; //field name -> multiple attributes (attributes have name and value)

        public static class FieldInfo {
            public enum FieldType { URN, URL, UID, STRING, DATETIME, EMAIL, KEY, BOOLEAN, CERTIFICATE, CREDENTIALS, INTEGER,
                PEERS/*Vector of Hastables containing "url" and "version"*/
            }

            public enum CreationAllowed { REQUIRED, ALLOWED, NOT_ALLOWED }

            public enum Protect { PUBLIC, PRIVATE, IDENTIFYING }

            private Map<String, Object> attributes; //all attributes
            private Map<String, Object> extraAttributes; //only non recognized

            private final String name;
            //known fields with defaults
            private FieldType fieldType;
            private CreationAllowed creationAllowed = CreationAllowed.NOT_ALLOWED;
            private boolean match = true;
            private boolean updateable = false;
            private Protect protect = Protect.PUBLIC; //default
            private String object = null;

            @Override
            public String toString() {
                return "FieldInfo{" +
                        "name=" + name +
                        ", fieldType=" + fieldType +
                        ", creationAllowed=" + creationAllowed +
                        ", match=" + match +
                        ", updateable=" + updateable +
                        ", protect=" + protect +
                        ", object='" + object + '\'' +
                        '}';
            }

            public FieldInfo(String name, FieldType fieldType) {
                this.name = name;
                this.fieldType = fieldType;
                //TODO initialise and fill in attributes and extraAttributes
            }
            public FieldInfo(String object, String name, FieldType fieldType, CreationAllowed creationAllowed, boolean match, boolean updateable) {
                this.name = name;
                this.fieldType = fieldType;
                this.creationAllowed = creationAllowed;
                this.match = match;
                this.updateable = updateable;
                this.object = object;
                //TODO initialise and fill in attributes and extraAttributes
            }
            public FieldInfo(String object, String name, FieldType fieldType, boolean match, Protect protect) {
                this.name = name;
                this.fieldType = fieldType;
                this.match = match;
                this.protect = protect;
                this.object = object;
                //TODO initialise and fill in attributes and extraAttributes
            }
            public FieldInfo(String object, String name, FieldType fieldType, CreationAllowed creationAllowed, boolean match, boolean updateable, Protect protect) {
                this.name = name;
                this.fieldType = fieldType;
                this.creationAllowed = creationAllowed;
                this.match = match;
                this.updateable = updateable;
                this.protect = protect;
                this.object = object;
                //TODO initialise and fill in attributes and extraAttributes
            }

            public FieldInfo(String name, Hashtable<String, Object> attributesTable) throws BadReplyGeniException {
                attributes = new HashMap<String, Object>();
                extraAttributes = new HashMap<String, Object>();

                this.name = name;
                for (Map.Entry<String, Object> entry : attributesTable.entrySet()) {
                    boolean known = false;

                    assert entry != null;
                    assert entry.getKey() != null;
                    assert entry.getValue() != null;

                    if (entry.getKey().equals("TYPE")) {
                        try {
                            fieldType = FieldType.valueOf((String) entry.getValue()); //throws exception when unknown value
                        } catch (IllegalArgumentException e) {
                            throw new BadReplyGeniException("field TYPE with value \""+entry.getValue()+"\" is unknown");
                        }
                        if (fieldType == null)
                            throw new BadReplyGeniException("field TYPE with value \""+entry.getValue()+"\" is unknown");
                        known = true;
                    }

                    if (entry.getKey().equals("CREATE")) {
                        try {
                            creationAllowed = CreationAllowed.valueOf(((String) entry.getValue()).replace(' ', '_')); //throws exception when unknown value
                        } catch (IllegalArgumentException e) {
                            throw new BadReplyGeniException("field CREATE with value \""+entry.getValue()+"\" is unknown");
                        }
                        if (creationAllowed == null)
                            throw new BadReplyGeniException("field CREATE with value \""+entry.getValue()+"\" is unknown");
                        known = true;
                    }

                    if (entry.getKey().equals("UPDATE")) {
                        if (entry.getValue() instanceof String && ((String) entry.getValue()).equalsIgnoreCase("true")) {
                            updateable = true;
                            known = true;
                        }
                        if (entry.getValue() instanceof Boolean) {
                            updateable = (Boolean) entry.getValue();
                            known = true;
                        }
                        if (entry.getValue() instanceof String && ((String) entry.getValue()).equalsIgnoreCase("false")) {
                            updateable = false;
                            known = true;
                        }
                        if (!known)
                            throw new BadReplyGeniException("field UPDATE with value \""+entry.getValue()+"\" is unknown");
                    }

                    if (entry.getKey().equals("MATCH")) {
                        if (entry.getValue() instanceof String && ((String) entry.getValue()).equalsIgnoreCase("true")) {
                            match = true;
                            known = true;
                        }
                        if (entry.getValue() instanceof Boolean) {
                            match = (Boolean) entry.getValue();
                            known = true;
                        }
                        if (entry.getValue() instanceof String && ((String) entry.getValue()).equalsIgnoreCase("false")) {
                            match = false;
                            known = true;
                        }
                        if (!known)
                            throw new BadReplyGeniException("field MATCH with value \""+entry.getValue()+"\" is unknown");
                    }

                    if (entry.getKey().equals("PROTECT")) {
                        try {
                            protect = Protect.valueOf((String) entry.getValue()); //throws exception when unknown value
                        } catch (IllegalArgumentException e) {
                            throw new BadReplyGeniException("field PROTECT with value \""+entry.getValue()+"\" is unknown");
                        }
                        if (protect == null)
                            throw new BadReplyGeniException("field PROTECT with value \""+entry.getValue()+"\" is unknown");
                        known = true;
                    }

                    if (entry.getKey().equals("OBJECT")) {
                        object = (String) entry.getValue();
                        known = true;
                    }

                    if (!known)
                        extraAttributes.put(entry.getKey(), entry.getValue());
                    attributes.put(entry.getKey(), entry.getValue());
                }


            }

            /** query all attributes, including the known ones (TYPE, CREATE, UPDATE, PROTECT). */
            public Object getAttribute(String name) {
                return attributes.get(name);
            }

            /** query only unknown attributes, excluding the known ones (TYPE, CREATE, UPDATE, PROTECT). */
            public Object getExtraAttribute(String name) {
                return extraAttributes.get(name);
            }

            public FieldType getType() {
                return fieldType;
            }

            public CreationAllowed getCreationAllowed() {
                return creationAllowed;
            }

            public boolean isUpdateable() {
                return updateable;
            }

            public Protect getProtect() {
                return protect;
            }

            public String getName() {
                return name;
            }

            public boolean isMatchable() {
                return match;
            }

            /**
             * @return null if OBJECT not specified. In that case it is the default for the API
             *   (i.e. SLICE for Slice Authority, MEMBER for Member Authority, SERVICE for Clearinghouse) */
            public String getObject() {
                return object;
            }
        }

        public GetVersionResult(Hashtable<String, Object> versionInfo, boolean v2) throws BadReplyGeniException {
            version = apiSpecifiesNonNullString(versionInfo.get("VERSION"));

            if (versionInfo.containsKey("CREDENTIAL_TYPES")) {
                if (!v2) {
                    Vector<String> strings = apiSpecifiesVectorOfString(versionInfo.get("CREDENTIAL_TYPES"));
                    supportedCredentialTypes = new ArrayList<CredentialType>();
                    for (String s : strings) {
                        supportedCredentialTypes.add(new CredentialType(s, "unknown"));
                    }
                } else {
                    Vector<Hashtable> credType = apiSpecifiesVectorOfT(Hashtable.class, versionInfo.get("CREDENTIAL_TYPES"));
                    supportedCredentialTypes = new ArrayList<CredentialType>();
                    for (Hashtable ht : credType) {
                        CredentialType credentialType = new CredentialType(
                                apiSpecifiesNonNullString(ht.get("type")),
                                apiSpecifiesNonNullString(ht.get("version"))
                        );
                        supportedCredentialTypes.add(credentialType);
                    }
                }
            }

            fieldInfos = new HashMap<String, FieldInfo>(); //always initialize, even if FIELDS not specified
            if (versionInfo.containsKey("FIELDS")) {
                Hashtable<String, Object> fields = apiSpecifiesHashtableStringToObject(versionInfo.get("FIELDS"));
                for (Map.Entry<String, Object> entry : fields.entrySet()) {
                    FieldInfo fieldInfo = new FieldInfo(entry.getKey(), apiSpecifiesHashtableStringToObject(entry.getValue()));
                    fieldInfos.put(entry.getKey(), fieldInfo);
                }
            }
        }

        public String getVersion() {
            return version;
        }

        public List<CredentialType> getSupportedCredentialTypes() {
            return supportedCredentialTypes;
        }

        public Map<String, FieldInfo> getFields() {
            return fieldInfos;
        }

        public Map<String, FieldInfo> getFieldsForObject(String objectName) {
            if (fieldInfos == null) return null;
            Map<String, FieldInfo> objectFields = new HashMap<String, FieldInfo>();
            for (Map.Entry<String, FieldInfo> e : fieldInfos.entrySet()) {
                String fieldName = e.getKey();
                FieldInfo fieldInfo = e.getValue();
                if (fieldInfo.getObject() != null && objectName != null && fieldInfo.getObject().equals(objectName))
                    objectFields.put(fieldName, fieldInfo);
            }
            return objectFields;
        }

        @Override
        public String toString() {
            return "GetVersionResult{" +
                    "version='" + version + '\'' +
                    ", supportedCredentialTypes=" + supportedCredentialTypes +
                    ", fieldInfos=" + fieldInfos +
                    '}';
        }
    }

    public static class FederationApiReply<T> implements SfaApiCallReply<T> {
        private final GeniUniformFederationApiResponseCode genicode;
        private final T val;
        /* output is typically set only on error*/
        private final String output;

        private final Hashtable rawResult;

        private final XMLRPCCallDetailsWithCodeValueError xmlRpcCallDetailsWithCodeValueError;

        public int getCode() {
            return genicode.getCode();
        }
        public GeniUniformFederationApiResponseCode getGeniResponseCode() {
            return genicode;
        }

        public T getValue() {
            return val;
        }

        /** can be null */
        public String getOutput() {
            return output;
        }

        public XMLRPCCallDetailsWithCodeValueError getXmlRpcCallDetailsWithCodeValueError() {
            return xmlRpcCallDetailsWithCodeValueError;
        }

        public static boolean isSuccess(XMLRPCCallDetails res) {
            if (res instanceof XMLRPCCallDetailsWithCodeValueError) {
                XMLRPCCallDetailsWithCodeValueError geniDetails = (XMLRPCCallDetailsWithCodeValueError) res;
                if (!(geniDetails.getResultCode() instanceof Integer))
                    return false;
                int code = (Integer) geniDetails.getResultCode();
                GeniResponseCode genicode;
                try {
                    genicode = GeniUniformFederationApiResponseCode.getByCode(code);
                } catch (UnknownResponseCodeException e) {
                    LOG.error("UnknownResponseCodeException: \""+e.getMessage()+"\" -> will be treated as unsuccessful", e);
                    genicode = GeniUniformFederationApiResponseCode.SERVER_REPLY_ERROR;
                }
                return genicode.isSuccess();
            }
            return false; //not a geni reply, so not successful
        }

        private static <TT> TT findVal(XMLRPCCallDetailsWithCodeValueError res) {
            Hashtable r = res.getResult();
            try {
                //noinspection unchecked
                return (TT) r.get("value");
            } catch (/*ClassCast*/Exception e) {
                //ignore
                return null;
            }
        }
        public FederationApiReply(XMLRPCCallDetailsWithCodeValueError res) {
            this(res, FederationApiReply.<T>findVal(res));
        }

        public FederationApiReply(XMLRPCCallDetailsWithCodeValueError res, T val) {
            this.xmlRpcCallDetailsWithCodeValueError = res;
            this.rawResult = res.getResult();

            this.val = val;
            Object codeObject = res.getResultCode();
            int code;
            if (codeObject != null) {
                if (codeObject instanceof Integer) {
                    code = (Integer) codeObject;
                } else {
                    LOG.error("Reply parse error: codeObject is not Integer but "+codeObject.getClass().getName());
                    code = GeniUniformFederationApiResponseCode.SERVER_REPLY_ERROR.getCode();
                }
            } else {
                LOG.error("Reply parse error: codeObject=null");
                code = GeniUniformFederationApiResponseCode.SERVER_REPLY_ERROR.getCode();
            }
            GeniUniformFederationApiResponseCode c;
            try {
                c = GeniUniformFederationApiResponseCode.getByCode(code);
            } catch (UnknownResponseCodeException e) {
                LOG.error("UnknowResponseCodeException: \""+e.getMessage()+"\"  codeObject="+codeObject, e);
                c = GeniUniformFederationApiResponseCode.SERVER_REPLY_ERROR;
            }
            this.genicode = c;

            String o = null;
            Hashtable r = res.getResult();
            if (r != null && r.get("output") != null) {
                o = r.get("output").toString();
                //any empty structure is interpreted as an empty string
                if (r.get("output") instanceof Hashtable && ((Hashtable)r.get("output")).isEmpty() )
                    o = "";
                if (r.get("output") instanceof Vector && ((Vector)r.get("output")).isEmpty() )
                    o = "";
            }
            this.output = o; //this should be a string, but we allow more. This may be null on success.
        }

        public Hashtable getRawResult() {
            return rawResult;
        }
        public Object getRawValue() {
            if (rawResult == null) return null;
            return rawResult.get("value");
        }

        @Override public XMLRPCCallDetailsWithCodeValueError getXMLRPCCallDetailsWithCodeValueError() { return xmlRpcCallDetailsWithCodeValueError; }


        @Override public XMLRPCCallDetails getXMLRPCCallDetails() { return xmlRpcCallDetailsWithCodeValueError; }
    }

    public static boolean isBusyApiReply(be.iminds.ilabt.jfed.lowlevel.XMLRPCCallDetails res) {
        if (res == null) return false;
        if (res instanceof XMLRPCCallDetailsWithCodeValueError) {
            XMLRPCCallDetailsWithCodeValueError geniDetails = (XMLRPCCallDetailsWithCodeValueError) res;

            if (geniDetails.getResultCode() == null)
                return false;

            int code = GeniAMResponseCode.GENIRESPONSE_SUCCESS.getCode();
            if (geniDetails.getResultCode() instanceof Hashtable) {
                Hashtable codeStruct = (Hashtable) geniDetails.getResultCode();
                code = (Integer) codeStruct.get("geni_code");
            }
            if (geniDetails.getResultCode() instanceof Integer) {
                code = (Integer) geniDetails.getResultCode();
            }

            if (GeniAMResponseCode.getByCode(code).isBusy())
                return true;
        }
        return false;
    }

    @Override
    protected boolean isBusyReply(be.iminds.ilabt.jfed.lowlevel.XMLRPCCallDetails res) {
        return isBusyApiReply(res);
    }


    public AbstractFederationApi(Logger logger, RetrySettings retrySettings, ApiInfo.Api serverType, JFedPreferences jFedPreferences) {
        super(logger, retrySettings, serverType, jFedPreferences);
    }


    public static List<AnyCredential> apiSpecifiesVectorOfCredentials(String credentialBaseName, Object o) throws BadReplyGeniException, CredentialException {
        Vector<Hashtable> vectOfDicts = AbstractApi.apiSpecifiesVectorOfT(Hashtable.class, o);

        int i = 0;

        List<AnyCredential> res = new ArrayList<AnyCredential>();
        for (Hashtable ht : vectOfDicts) {

            String credName = credentialBaseName;
            if (vectOfDicts.size() > 1)
                credName += " "+i;

            AnyCredential cred = AnyCredential.createFromV3Hashtable(credName, ht);

            //assert below is useful test during debugging, but off course not to be assumed in all cases!
//            assert cred instanceof SfaCredential : "credential is not SfaCredential but "+cred.getClass().getName()+" ht="+ht;

            res.add(cred);

            i++;
        }

        return res;
    }

    /**
     * The result of a lookup operation.
     *
     * This is indexed in 2 ways:
     *   - by urn (or ID) (String or GeniUrn)
     *   - by index nr
     *
     *   Note: some lookups (for KEY for example) will not use URNs! Be aware of that.
     *
     * The objects returned are accessable as either:
     *  - Hashtable<String, Object>
     *  - an object of the class "T"
     * */
    public static class LookupResult<T> {
        Hashtable<String, Hashtable<String, Object>> raw;
        private List<String> keys;
        private LookupResultConverter<T> converter;

        public LookupResult(Object res, LookupResultConverter<T> converter) throws BadReplyGeniException {
            this.converter = converter;
            Hashtable<String, Object> res2 = apiSpecifiesHashtableStringToObject(res);
            raw = new Hashtable<String, Hashtable<String, Object>>();
            for (Map.Entry<String, Object> o : res2.entrySet()) {
                if (o.getValue() == null) {
                    throw new BadReplyGeniException("expecting result to lookup to be a mapping of Strings (urn or other ID) to Dictionaries of field/value pairs. However, instead of a dictionary, a null was found for key \""+o.getKey()+"\"!");
                }
                if (o.getValue() instanceof Vector) {
                    throw new BadReplyGeniException("expecting result to lookup to be a mapping of Strings (urn or other ID) to Dictionaries of field/value pairs. However, instead of a dictionary, a List was found for key \""+o.getKey()+"\"!");
                }
                if (!(o.getValue() instanceof Hashtable)) {
                    throw new BadReplyGeniException("expecting result to lookup to be a mapping of Strings (urn or other ID) to Dictionaries of field/value pairs. However, instead of a dictionary, a "+(o.getValue().getClass().getName())+" was found for key \""+o.getKey()+"\"!");
                }
                Hashtable<String, Object> o2 = apiSpecifiesHashtableStringToObject(o.getValue());
                raw.put(o.getKey(), o2);
            }
            keys = new ArrayList<String>(raw.keySet());
            assert keys.size() == raw.size();
        }

        public List<String> getKeys() {
            return Collections.unmodifiableList(keys);
        }

        public int size() {
            return raw.size();
        }

        public boolean isEmpty() {
            return raw.isEmpty();
        }

        public String getKey(int i) {
            return keys.get(i);
        }
        public Hashtable<String, Object> get(int i) {
            return raw.get(keys.get(i));
        }

        public Hashtable<String, Object> get(String urn) {
            return raw.get(urn);
        }
        public Hashtable<String, Object> get(GeniUrn urn) {
            return raw.get(urn.getValue());
        }

        public T getConverted(int i) throws JFedException {
            if (convertedList != null) return convertedList.get(i); //use cache if available
            if (converter == null) throw new RuntimeException("No LookupResultConverter specified");
            String urn = keys.get(i);
            Hashtable<String, Object> val = raw.get(urn);
            return converter.convert(urn, val);
        }

        public T getConverted(String urn) throws JFedException {
            if (converter == null) throw new RuntimeException("No LookupResultConverter specified");
            Hashtable<String, Object> val = raw.get(urn);
            return converter.convert(urn, val);
        }

        public T getConverted(GeniUrn urn) throws JFedException {
            if (converter == null) throw new RuntimeException("No LookupResultConverter specified");
            Hashtable<String, Object> val = raw.get(urn.getValue());
            return converter.convert(urn.getValue(), val);
        }

        private List<T> convertedList; //cached version of list: conversion is only done once
        public List<T> getConvertedList() throws JFedException {
            if (convertedList != null) return convertedList;

            List<T> res = new ArrayList<T>(raw.size());
            for (Map.Entry<String, Hashtable<String, Object>> e :  raw.entrySet()) {
                T t = converter.convert(e.getKey(), e.getValue());
                res.add(t);
            }
            convertedList = Collections.unmodifiableList(res);
            return convertedList;
        }

        public Hashtable<String, Hashtable<String, Object>> getRaw() {
            return raw;
        }

        @Override
        public String toString() {
            return "LookupResult{" +
                    "" + raw +
                    '}';
        }
    }

    public interface LookupResultConverter<T> {
        T convert(String urn, Hashtable<String, Object> o) throws JFedException;
    }


    public abstract List<GetVersionResult.FieldInfo> getMinimumFields(String objectName);
    public List<String> getMinimumFieldNames(String objectName) {
        assert objectName != null;
        List<String> res = new ArrayList<String>();
        List<GetVersionResult.FieldInfo> minFields = getMinimumFields(objectName);
        if (minFields == null) return null;
        for (GetVersionResult.FieldInfo f :  minFields)
            res.add(f.getName());
        return res;
    }
    public Map<String, String> getMinimumFieldsMap(String objectName) {
        assert objectName != null;
        Map<String, String> res = new HashMap<String, String>();
        List<String> minFieldNames = getMinimumFieldNames(objectName);
        if (minFieldNames == null) return null;
        for (String key : minFieldNames) {
            res.put(key, key+" value");
        }
        return res;
    }
    public abstract List<String> getApiObjects();
    public abstract List<String> getOptionalApiServices();
    public abstract List<String> getRequiredApiServices();


    public void setHandleMalformedReplies(boolean handleErrors) {
        this.handleMalformedReplies = handleErrors;
    }
}
